# Copyright (C) 2004 Philippe Normand <phil@respyre.org>
#
# This file is part of XMLObject (http://xmlobject.base-art.net)
#
# Under PSF License (see COPYING)
# $HeadURL: https://svn.base-art.net/full/xmlobject/trunk/XMLObject/Attributes.py $
# $Id: Attributes.py 19 2004-08-15 15:47:30Z test $


import classregistry, utils
from Node import Node, RequiredNodeError

__all__ = [ 'CDATAttribute', 'NMTokenAttribute', 'NMTokensAttribute',
            'IntegerAttribute', 'StringAttribute' ]

class Attribute(Node):
    """ Base Class for Attributes

        This should not be directly used. It's only a general class
        used by the 'real' attribute implementations.
    """
    
    def __init__(self, default='', type='Attribute', permittedValues=None, optional=False):
        if optional:
            default = None
        # set noLimit to False for all attributes.
        # This isn't negociable :) Why would you want multiple
        # attributes of the same name in one XML element ?
        Node.__init__(self, type=type, noLimit = False,
                      default=default, optional=optional)
        self.permittedValues = permittedValues

    def checkType(self, val):
        result = True
        if self.permittedValues is not None:
            result = val in self.permittedValues
        return result

class CDATAttribute(Attribute):
    
    def checkType(self, val):
        result = False
        if self.isOptional() and val is None:
            result = True
        elif type(val) == type('') and Attribute.checkType(self, val):
            result = True
        return result

    def xmlrepr(self, parentInstance=None):
        val = self.getValue()
        if val is not None:
            return '%s="%s"' % (utils.lowWord(self.getName()),val)
        else:
            return ''

    def getValueFromDom(self, dom, attrName, **kw):
        xmlAttrNode = dom.getAttributeNode(utils.lowWord(attrName))
        try:
            value = utils.encode(xmlAttrNode.value)
        except AttributeError:
            if not self.isOptional():
                raise RequiredNodeError(self.getName())
            else:
                value = None
        return value

class NMTokenAttribute(CDATAttribute):

    _valids = [':','_','-','.']
    
    def checkType(self, val):
        result = CDATAttribute.checkType(self,val)
        if result and val is not None:
            tmpVal = val or ''
            for car in self._valids:
                tmpVal = tmpVal.replace(car,'')
            result = tmpVal.isalnum()
        return result


class NMTokensAttribute(NMTokenAttribute):

    _valids = NMTokenAttribute._valids + [' ', '\t', '\r\n', '\n']

    
class StringAttribute(CDATAttribute):
    """
        Suppose you have an XMLObject type like this one:

        class FooBar(XMLObject):
            glop = StringAttribute()

        FooBar(glop="Hey Yah").toXml() would output:

        <fooBar glop="Hey Yah" />
    """

    def __init__(self, default='', permittedValues=None, optional=False):
        if optional:
            default = None
        CDATAttribute.__init__(self, type='StringAttribute',
                               permittedValues=permittedValues,
                               default=default, optional=optional)

    def xmlrepr(self, parentInstance=None):
        val = self.getValue()
        if val is not None:
            val = utils.replaceAll(val, parentInstance.getEntities())
            return '%s="%s"' % (utils.lowWord(self.getName()),val)
        else:
            return ''

class IntegerAttribute(Attribute):
    """

    """
    
    def __init__(self, default=0, permittedValues=None, optional=False):
        if optional:
            default = None
        Attribute.__init__(self, type='IntegerAttribute',
                           permittedValues=permittedValues,
                           default=default, optional=optional)

    def xmlrepr(self, parentInstance=None):
        val = self.getValue()
        if val is not None:
            return '%s="%d"' % (utils.lowWord(self.getName()), val)
        else:
            return ''

    def getValueFromDom(self, dom, attrName, **kw):
        xmlAttrNode = dom.getAttributeNode(utils.lowWord(attrName))
        if not self.isOptional() and not xmlAttrNode:
            raise RequiredNodeError(self.getName())
        if not xmlAttrNode:
            value = None
        else:
            value = int(utils.encode(xmlAttrNode.value))
        return value

    def checkType(self, val):
        result = False
        if self.isOptional() and val is None:
            result = True
        elif type(val) == type(0) and Attribute.checkType(self, val):
            result = True
        return result
