# Copyright (C) 2004 Philippe Normand <phil@respyre.org>
#
# This file is part of XMLObject (http://xmlobject.base-art.net)
#
# Under PSF License (see COPYING)
# $HeadURL: https://svn.base-art.net/full/xmlobject/trunk/XMLObject/Nodes.py $
# $Id: Nodes.py 19 2004-08-15 15:47:30Z test $


import string, copy

from Node import Node, RequiredNodeError
from main import XMLObject
from TypedList import TypedList
from MixedList import MixedList
import classregistry, utils

__all__ = [ 'RawNode', 'TextNode', 'ItemNode', 'ChoiceNode',
            'ListNode', 'CommentNode' ]

class CommentNode(Node):
    """ Insert comments in the XML object

        CommentNodes are optional by default. May I change this in
        near future :)

        Example:

        class CommentTator(XMLObject):
            myComment = CommentNode()

        print CommentTator(myComment='Blah lbah').toXml(headers=0)

        <commentTator>
           <!-- Blah lbah -->
        </commentTator>
    """
    
    def __init__(self, default='', optional=True):
        Node.__init__(self, type='Comment', optional=optional,
                      default=default, noLimit=False)

    def xmlrepr(self, parentInstance=None):
        value = self.getValue()
        if value and parentInstance:
            value = utils.replaceAll(value, parentInstance.getEntities())
            if self.getName() != 'mainNode':
                value = "<!-- %s -->" % value
        else:
            value = ''
        return value
        
    def getValueFromDom(self, dom, attrName, **kw):
        value = ''
        for node in dom.childNodes:
            if node.nodeType == node.COMMENT_NODE:
                value += node.data
        value = utils.encode(value.strip())
        if value == '':
            value = None
        return value

class TextNode(Node):
    """ Example class:

        class Foo(XMLObject):
            text = TextNode()

        f = Foo(text = 'Kind of fun')
        print f.toXml(headers=False)
        
        Example XML output:

        <foo>
            <text>
               Kind of fun
            </text>
        </foo>
    """

    def __init__(self, default = None, optional=False):
        Node.__init__(self, type = 'Text', optional=optional,
                      default = default, noLimit=False)

    def xmlrepr(self, parentInstance=None):
        value = self.getValue()
        if value and parentInstance:
            value = utils.replaceAll(value, parentInstance.getEntities())
            if self.getName() != 'mainNode':
                value = "<%s>%s</%s>" % (self.getName(), value, self.getName())
        else:
            value = ''
        return value
    
    def getValueFromDom(self, dom, attrName, **kw):
        value = ''
	candidates = utils.getDirectChildrenWithName(dom,utils.lowWord(self.getName()))
        if len(candidates) == 0:
	    candidates = utils.getDirectChildrenWithName(dom,self.getName())
            if len(candidates) == 0:
                if self.getName() == 'mainNode':
                    candidates = [dom]
                elif not self.isOptional():
                    raise RequiredNodeError(self.getName(),dom.toxml())
        for candidate in candidates:
            for node in candidate.childNodes:
                if node.nodeType == node.TEXT_NODE:
                    value += node.data
        value = utils.encode(value.strip())
        if value == '':
            value = None
        return value

class RawNode(Node):
    """ Example class:

        class Foo(XMLObject):
            text = RawNode()

        f = Foo(text = 'Kind of fun')
        print f.toXml(headers=False)
        
        Example XML output:

        <foo>
             <![CDATA[Kind of fun]]>
        </foo>
    """

    def __init__(self, default = None, optional=False):
        Node.__init__(self, type = 'Raw', optional=optional,
                      default = default, noLimit=False)

    def xmlrepr(self, parentInstance=None):
        value = self.getValue()
        if not value:
            value = ''
        if value != '':
            value = "<![CDATA[%s]]>" % value
        return value
    
    def getValueFromDom(self, dom, attrName, **kw):
        value = ''
        for node in dom.childNodes:
            if node.nodeType == node.CDATA_SECTION_NODE:
                value += node.data
        value = utils.encode(value.strip())
        if value == '':
            value = None
        return value


class ItemNode(Node):
    """ Class used to refer to an existing XMLObject type.
        Suppose you have an XMLObject type like this one:

        class FooBar(XMLObject):
            glop = StringAttribute()

        You can then use it in other XMLObjects:

        class Blah(XMLObject):
            fb = ItemNode('FooBar')

        blah = Blah( fb = FooBar(glop = 'HelloWorld') )

        blah.toXml() would output:

            <blah>
                 <fooBar glop="HelloWorld" />
            </blah>

        `optional` keyword specifies wether the Item is mandatory or not.
    """

    def __init__(self, itemType, default=None, optional=False):
        Node.__init__(self, type='Item', itemType=itemType, noLimit=False,
                      optional=optional, default=default)

    def xmlrepr(self, parentInstance=None):
        val = self.getValue()
        if val is not None:
            return val.toXml(headers=0)
        else:
            return ''

    def getValueFromDom(self, dom, attrName, **kw):
        itemTypeName = self.getItemType()
        parentTypeName = utils.lowWord(self.getParentType())
        xmlObject = classregistry.registry(kw['registry']).getClass(itemTypeName)()
	candidates = utils.getDirectChildrenWithName(dom,utils.lowWord(itemTypeName))
        if len(candidates) == 0:
	    candidates = utils.getDirectChildrenWithName(dom,itemTypeName)
            parentTypeName = self.getParentType()
            if len(candidates) == 0 and not self.isOptional():
                raise RequiredNodeError(self.getName())
        value = None
        for candidate in candidates:
            if candidate.parentNode.nodeName == parentTypeName:
                value = xmlObject._fromDom(candidate)
                break
        return value

class ChoiceNode(Node):
    """ Alternative Nodes

        Right recursivity is permitted, but left recursions are (I hope so)
        prohibited. In the following example, `mix` can be an `Item` or
        a `Blah` object:
    
        class Item(XMLObject):
            pass

        class Blah(XMLObject):
            mix = ChoiceNode(['Item','Blah'])

        One can also use the special '#PCDATA' alternative:

        class Blah(XMLObject):
            mix = ChoiceNode(['#PCDATA', 'Item'])

        So `mix` can be either a string or an Item instance.
        `noLimit` keyword can be set to True when the node has to
        bahave as a List.
        Setting `optional` to True, XMLObject won't complain it it's empty.
    """

    def __init__(self, choiceAlternatives, optional=False, noLimit=False):
        self.alternatives = choiceAlternatives
        default = None
        if noLimit:
            default = MixedList(self)
        Node.__init__(self, type='Choice', itemType=choiceAlternatives,
                        optional=optional, default=default, noLimit=noLimit)

    def checkForLeftRecursivity(self, xmlObject, nodeName):
        if self.getParentType() == nodeName:
            raise ValueError('Left recursivity detected in "%s"' % self.getName())
        elif isinstance(xmlObject, XMLObject):
            for child, instance in xmlObject.getNodes().iteritems():
                exc = ValueError("""\
Left recursivity detected in "%s" through "%s" attribute of "%s" """ %
                                 (self.getName(), child, xmlObject.getName()))
                if instance.getItemType() == self.getParentType():
                    raise exc
                elif instance.getType() == 'Choice':
                    if instance.getItemType()[0] == self.getParentType():
                        raise exc

    def resetValue(self):
        if self.isNoLimit():
            self.setValue(MixedList(self))
        else:
            self.setValue(None)

    def setValue(self, value):
        if isinstance(value, ChoiceNode):
            value = copy.deepcopy(value)
        Node.setValue(self, value)
                        
    def checkType(self, val):

        if len(self.alternatives) > 0:
            self.checkForLeftRecursivity(val,self.alternatives[0])

        result = False
        if val is None:
            result = True
        elif isinstance(val, MixedList):
            dummyMixedList = MixedList(self)
            ok = True
            for it in val:
                try:
                    dummyMixedList.checkItem(it)
                except:
                    ok = False
                    break
            result = ok
        elif isinstance(val, XMLObject):
            result = val.getName() in self.alternatives
        elif '#PCDATA' in self.alternatives and type(val) == type(''):
            result = True
        return result

    def xmlrepr(self, parentInstance=None):
        val = self.getValue()
        result = ''
        if val is not None:
            if isinstance(val, MixedList):
                for node in val:
                    if isinstance(val, parentInstance.__class__):
                        result += node.toXml(headers=0)
                    else:
                        result += str(node)
            elif isinstance(val, parentInstance.__class__):
                result += val.toXml(headers=0)
            else:
                result += str(val)
        return result
    
    def getValueFromDom(self, dom, attrName, **kw):
        parentTypeName = self.getParentType()
        result = None
        if self.isNoLimit():
            result = MixedList(self)

        for childNode in dom.childNodes:
            for alt in self.alternatives:
                value = None
                if alt == '#PCDATA':
                    if childNode.nodeType == childNode.TEXT_NODE:
                        value = childNode.data
                        value = utils.encode(value.strip())                        
                elif childNode.nodeName == utils.lowWord(alt):
                    xmlObject = classregistry.registry(kw['registry']).getClass(alt)()
                    value = xmlObject._fromDom(childNode)
                if not value:
                    continue
                elif not self.isNoLimit():
                    return value
                else:
                    result.append(value)
        return result

        
class ListNode(Node):
    """ Single Typed List node

        ListNode is used to represent a variable number of XMLObjects,
        all of the same type. Example:

        class Apple(XMLObject):
            type =  StringAttribute()
            color = StringAttribute()
            
        class AppleBag(XMLObject):
            apples = ListNode('Apple')

        bag = AppleBag()
        bag.apples.append(Apple(type='Blah',color='green'))
        print bag.toXml(headers=0)
        
        <appleBag>
          <apple type="Blah" color="green" />
          <!-- other apples can fit here :) -->
        </appleBag>


        If `optional` flag set to False, the bag has to carry at least
        one apple. 
    """
    
    def __init__(self, itemType, optional=True):
        Node.__init__(self, type='List', itemType=itemType,
                      optional=optional, noLimit=True,
                      default = TypedList(self))
        
    def checkType(self, aList):
        result = True
        dummyTypedList = TypedList(self)
        itemsNb = len(aList)
        if itemsNb > 1 and not self.isNoLimit():
            result = False
        if result:
            for item in aList:
                try:
                    dummyTypedList.checkItem(item)
                except:
                    result = False
                    break
        return result

    def xmlrepr(self, parentInstance=None):
        result = ''
        itemsNb = len(self.getValue())
        if itemsNb < 1 and not self.isOptional():
            raise TypeError("Expected some node in '%s.%s' as it is not optional" %
                            (parentInstance.getName(),self.getName()))
        for node in self.getValue():
            nodeRepr = node.toXml(headers=0)
            if node.getName() != self.getItemType():
                result += "<%s>%s</%s>" % (utils.lowWord(self.getItemType()),
                                           nodeRepr,
                                           utils.lowWord(self.getItemType()))
            else:
                result += nodeRepr
        return result

    def resetValue(self):
        self.setValue(TypedList(self))

    def getValueFromDom(self, dom, attrName, **kw):
        xmlList = TypedList(self)
        itemTypeName = self.getItemType()
        parentTypeName = utils.lowWord(self.getParentType())
        klass = classregistry.registry(kw['registry']).getClass(itemTypeName)
	candidates = utils.getDirectChildrenWithName(dom,utils.lowWord(itemTypeName))
        if len(candidates) == 0:
            parentTypeName = self.getParentType()
	    candidates = utils.getDirectChildrenWithName(dom,itemTypeName)
            if len(candidates) == 0 and (not self.isOptional()):
                raise RequiredNodeError(self.getName(),dom.toxml())
        for xmlNode in candidates:
            if xmlNode.parentNode.nodeName == parentTypeName:
                xmlObject = klass()
                xmlList.append( xmlObject._fromDom(xmlNode) )
        return xmlList
