# Copyright (C) 2004 Philippe Normand <phil@respyre.org>
#
# This file is part of XMLObject (http://xmlobject.base-art.net)
#
# Under PSF License (see COPYING)
# $HeadURL: https://svn.base-art.net/full/xmlobject/trunk/XMLObject/main.py $
# $Id: main.py 19 2004-08-15 15:47:30Z test $


# xml.minidom is only used to parse incoming XML
# when building XMLObject from string data

from xml.dom.minidom import parseString
import sys,re,types

import classregistry
import utils
from Node import Node
from Attributes import Attribute
from PrettyXMLPrinter import PrettyXMLPrinter, NoCallbackError

__all__ = [ 'XMLObject', 'ParseError' ]

class ParseError(Exception):

    def __init__(self, domException, xmlData):
        self._domExcp = domException
        self._xmlData = xmlData

    def __str__(self):
        return str(self._domExcp) + ':\n' + str(self._xmlData)

class MetaAttribute(object):
    """ Each XMLObject type deals with MetaAttributes. In the example below,
        `content` and `title` are MetaAttributes::

          >>> class MyXMLObject(XMLObject):
          ...   content = TextNode('Content') # _XO_content
          ...   title = StringAttribute()     # _XO_title
          ...
          >>>
          
        MetaAttributes handle value with the XMLObject they're included in.
        Type checking is done in `__set__`.
    """
    
    def __init__(self, nodeInstance):
        self._attrName = '_XO_%s' % nodeInstance.getName()
        self._node = nodeInstance
        
    def __get__(self, inst, cls):
        return inst.get(self._attrName)

    def __set__(self, inst, val):
        if self._node.checkType(val):
            inst.set(self._attrName,val)
        else:
            raise TypeError('Incorrect type in assignment of '+ `self._node.getName()`
                            + ' (%s)' % repr(val))
        
class MetaXMLObject(type):
    """ The place where all magic happens:

        - remember the name of the new type (class)
        - store attribute instances (used in XMLObject.{to,from}Xml methods)
        - create MetaAttributes
        - handle inheritance

        Special variables are:

        - `__name__` : the class name
        - `__nodes__` : the hash of class members keyed by name. Each node has
          to be an instance of the Node class.
    """

    __methods__ = [ '_isMember', 'set','get', 'getEntities',
                    'toXml', 'fromXml', '_fromDom','getName', 'getChildren']

    def __new__(cls, className, bases, dictionnary):

        # check if user defined class doesn't override esssential stuff
        intersection = utils.contains(dictionnary.keys(), cls.__methods__)
        if len(intersection) > 0 and className != 'XMLObject':
            raise TypeError('Unexpected methods defined in %s : %s' %
                            (className, str(intersection)))
        
        dictionnary['__name__'] = className
        dictionnary['__nodes__'] = {}

        nodes = filter(lambda x: x is not None,
                       map(lambda y,z: cls.isXONode(y,z),
                           dictionnary.keys(), dictionnary.values()))

        # build the MetaAttributes corresponding to the nodes of the class
        for name, instance in nodes:
            # store the attribute name somewhere
            instance.setName(name)
            if instance.getType() in ['List','Item','Choice']:
                instance.setParentType(className)
            dictionnary['__nodes__'][name] = instance
            dictionnary[name] = MetaAttribute(instance)

        newClass = type.__new__(cls, className, bases, dictionnary)

        # mutiple inheritance handling
        if cls.mro(newClass)[-1] != object:
            raise TypeError("Multiple inheritance is EVIL. Don't use it")
        
        # simple inheritance handling
        for superClass in cls.mro(newClass)[1:-2]:
            for nodeName,node in superClass.__nodes__.iteritems():
                newClass.__nodes__[nodeName] = node
                setattr(newClass, nodeName, MetaAttribute(node))

        # finally register the new class type so that it can be found
        # using its name
        classregistry.registry(newClass._registry).addClass(newClass)

        # propagate the registry id in each node 
        for nodeName,node in newClass.__nodes__.iteritems():
            node.setRegistry(newClass._registry)
            
        return newClass

    def isXONode(cls, name, instance):
        """ Checks if an instance variable is an XMLObject Node

        """
        result = None
        if isinstance(instance,Node) and '_' not in name[:1] \
           and name not in cls.__methods__ and type(instance) != types.FunctionType:
            result = (name,instance)
        return result
    isXONode = classmethod( isXONode )


class XMLObject(object):
    """
        Class attributes of interest:

        - `_name` : override `__name__`
        - `_entities` : user-defined entities stored in a dictionnary

        If you name one node with 'mainNode' string, this Node will be hooked so that
        calling Node methods ('append' for instance) from the XMLObject instance, the
        
        class Other(XMLObject):
            boo = StringAttribute(default='titi')

        class Single(XMLObject):
            mainNode = ChoiceNode(['Other','Single'], noLimit=True)
            title = StringAttribute()

        single = Single()
        single.title = 'Foo'
        single.append(Other())
        single.append(Other())
        single.append(single)

        <single title="Foo">
          <other boo="titi"/>
          <other boo="titi"/>
          <single title="Foo">
            <other boo="titi"/>
            <other boo="titi"/>
          </single>
        </single>
    """
    
    __metaclass__ = MetaXMLObject

    _registry = 'registryID'

    _defaultEntities = [ ('&','&amp;'),
                         ('<', '&lt;'),
                         ('>' , '&gt;'),
                         ('"' , '&quot;'),
                         ("'", '&apos;'),
                       ]

    _entities = [ ]

    _dtd = None

    def __init__(self, *args, **kw):
        self._attributes = {}

        self._prettyPrinter = PrettyXMLPrinter()
        
        # set default values for the new class instance
        for attrName, instance in self.getNodes().iteritems():
            instance.resetValue()
            setattr(self, attrName, instance.getValue())

        if len(args) == 1 and len(self.getChildren()) == 1:
            setattr(self, self.getChildren()[0],args[0])

        # set user values for the class instance
        # Some may override defaults
        for k,v in kw.iteritems():
            if self._isMember(k):
                setattr(self, k, v)

    #####################################################################
    ### Private
    #####################################################################
                
    def _isMember(self, name):
        " Dummy method to drop later "
        result = filter(lambda i: i == name, self.getNodes().keys())
        return len(result) > 0

    def _fromDom(self, dom, dummy=False,parentDom=None):
        """ Building an XMLObject given its pendant DOM tree.

        """
        for attrName, node in self.getNodes().iteritems():
            try:
                value = node.getValueFromDom(dom, attrName,
                                             registry=self._registry)
            except Exception, e:
                raise
            else:
                setattr(self, attrName, value)
        return self

    #####################################################################
    ### Nodes Access (Reserved to MetaAttribute class)
    #####################################################################
                
    def get(self, nodeName):
        """ Fetch a node value given its name.
            using directly (or implicitely) `getattr` is smarter.
            This method is used by the MetaAttribute instances
        """
        return self._attributes[nodeName]

    def set(self, nodeName, val):
        """ Modify a node value given its name.
            Same external behaviour as `setattr`
            This method is used by the MetaAttribute instances
        """
        self._attributes[nodeName] = val
        
    #####################################################################
    ### Hooks
    #####################################################################

    def __cmp__(self, other):
        """ Two XMLObjects are equal if they have the same XML representation
        """
        o1 = self.toXml(headers=0)
        o2 = other.toXml(headers=0)
        return cmp(o1, o2)

    def __getattr__(self, attrName):
        if '_' not in attrName[:1]:
            nodeName = 'mainNode'
            try:
                node = self.__nodes__[nodeName]
            except KeyError:
                raise AttributeError, """\
Hook through "%s" failed. Check you correctly declared this node.
Access to "%s" refused.""" % (nodeName,attrName)
            else:
                attr = MetaAttribute(node)
                return getattr(attr.__get__(self, XMLObject), attrName)
        raise AttributeError, attrName

    def __str__(self):
        return self.toXml(headers=0)

    #####################################################################
    ### 
    #####################################################################

    def getValue(self):
        return getattr(self, self.getName()+ '__')

    def getEntities(self):
        """ Get all entities specified in the xmlobject:

        - `_defaultEntities` stores the most common entities (<, >, ...)
        - `_entities` stores user-defined entities

        This method merges the two entities lists.
        """
        return self._entities + self._defaultEntities

    def setName(self, newName):
        self._name = newName

    def getName(self):
        """ Fetch XMLObject's instance name
            which can be redefined by `_name` class attribute
        """
        name = self.__name__
        if hasattr(self, '_name'):
            if type(self._name) != type(''):
                raise ValueError("Incorrect '_name' for %s: '%s'" % (name,self._name))
            name = self._name
        if ' ' in name:
            name = name.replace(' ','_')
        return name

    def getChildren(self):
        """ Basic Introspection

            Fetch XMLObject node names.
        """
        return self.getNodes().keys()

    def getNodes(self):
        """ Sub-Nodes accessor

            Fetch XMLObject nodes. Return a hash.
        """
        return self.__nodes__
        
    #####################################################################
    ### XML input/output
    #####################################################################

    def toXml(self, headers = 1, tabLength=2, encoding="iso-8859-1"):
        """ Exporting an XMLObject instance to XML.

            Optionnal keyword parameters are:
            - `headers` : boolean to figure out if XML data has to be headed by
              the <?xml ?> processing instruction
            - `tabLength`: XML indentation length
            - `encoding` : default charset to put in option <?xml?> PI.

            Return a string representing the XMLObject
        """
        entities = self.getEntities()
        result = ''
        if headers:
            result = '<?xml version="1.0" encoding="%s" ?>' % encoding
        attrs, others = '', ''
        name = self.getName()
        name = utils.lowWord(name)
        result += '<%s' % name

        mainNode = None
        for attrName, node in self.getNodes().iteritems():
            # 0. backup default node value
            default = node.getValue()
            # 1. set node value temporary
            node.setValue(getattr(self, attrName))
            nodeRepr = node.xmlrepr(parentInstance=self)
            if isinstance(node, Attribute):
                rpr = nodeRepr
                if nodeRepr:
                    attrs += ' ' + nodeRepr
            else:
                others += nodeRepr
            if attrName == 'mainNode':
                mainNode = nodeRepr
            # 2. reset value to default
            node.setValue(default)
        if attrs:
            result += attrs
        if others != '':
            if mainNode:
                try:
                    others = re.match('<mainNode>(.*)</mainNode>',mainNode).groups()[0]
                except:
                    others = mainNode
            result = "%s>%s</%s>" % (result, others, name)
        else:
            result = '%s/>' % result
        return self._prettyPrinter.prettyPrint(result, indent=' '*tabLength)

    def fromXml(cls, xmlData):
        """ Feeding an XMLObject instance with XML string data.

            `xmlData` is parsed by `xml.dom.minidom.parseString`.
            The resulting DOM tree is then used to build Nodes
            recursively.
        """
        val = cls()
        try:
            dom = parseString(xmlData)
        except Exception,ex:
            raise ParseError(ex, xmlData)
        else:
            val._fromDom(dom.documentElement)
        dom.unlink()
        return val

    fromXml = classmethod(fromXml)
