# Copyright (C) 2004 Philippe Normand <phil@respyre.org>
#
# This file is part of XMLObject (http://xmlobject.base-art.net)
#
# Under PSF License (see COPYING)

"""
    XMLObject's unit-tests

"""

import sys
sys.path.append('..')

from XMLObject import *
import unittest

False = 0!=0
True  = 0==0

class Item(XMLObject):
    _entities = [ ('&xml;', 'eXtensible Markup Language')]
    record = StringAttribute()
    position = IntegerAttribute(optional=False)
    content = TextNode()
    cdataAttr = CDATAttribute(permittedValues=['ok1','ok2'], optional=True)
    nmTok = NMTokenAttribute(optional=True)
    nmToks = NMTokensAttribute(optional=True)
    dummyData = RawNode()
    comment = CommentNode()
    
    def aMethod(self):
        return 'ok'

class Content(XMLObject):
    mainNode = TextNode()

class GetSetTest(unittest.TestCase):

    data = {'record':'AB1',
            'position':1,
            'content':'hey this is some content < &xml;',
            'dummyData': 'This is Dummy',
            'comment': 'this should be a comment'}

    xmlData = """\
<item record="AB1" position="1">
    <!-- this should be a comment -->
  <content>
      hey this is some content &lt; eXtensible Markup Language
  </content>
    <![CDATA[This is Dummy]]>
</item>"""

    def setUp(self):
        self.obj = Item(**self.data)

    def testGet(self):
        self.assertEqual(self.obj.content,'hey this is some content < &xml;')
        self.assertEqual(self.obj.position,1)
        self.assertEqual(self.obj.record, 'AB1')
        self.assertEqual(self.obj.dummyData, 'This is Dummy')
        self.assertEqual(self.obj.comment,'this should be a comment')
        
    def testSet(self):
        self.obj.position = 4
        self.assertEqual(self.obj.position, 4)
        # really dummy :( Could use self.obj.set() instead  of setattr()
        self.assertRaises(TypeError, setattr, self.obj, 'position', 'not a number')
        self.assertRaises(TypeError, setattr, self.obj, 'record', 2)

    def testNewAttrs(self):
        self.obj.cdataAttr = 'ok1'
        self.obj.nmTok = 'some_data:'
        s1 = self.obj.nmToks = """\
        few\t
lines of : text"""
        self.assertRaises(TypeError, setattr, self.obj, 'cdataAttr', '#not > ok')
        self.assertRaises(TypeError, setattr, self.obj, 'nmTok', 'few lines\n')
        self.assertRaises(TypeError, setattr, self.obj, 'nmToks', '#blah ')
        self.assertEqual(self.obj.nmTok,'some_data:')
        self.assertEqual(self.obj.cdataAttr,'ok1')
        self.assertEqual(self.obj.nmToks,s1)

                         
class Playlist(XMLObject):
    name = StringAttribute()
    type = StringAttribute()
    items = ListNode('Item')
    comment = CommentNode()
    
class GetSetTest2(unittest.TestCase):

    data = {'name':'foo',
            'comment': 'blah blah',
            'type':'xml' }

    def setUp(self):
        self.obj = Playlist(**self.data)
        for i in range(10):
            self.obj.items.append(Item(position=i, content=chr(i)*5))
        
    def testGetList(self):
        self.assertEqual(len(self.obj.items), 10)
        for i in range(10):
            self.assertEqual(self.obj.items[i].position,i)

    def testSetList(self):
        class DummyItem(XMLObject): pass
        self.assertRaises(TypeError, self.obj.items.append, DummyItem())
        it = Item(position=0, content='Foo Bar')
        self.obj.items = [ it ]
        self.assertEqual(self.obj.items[0], it)

class NewNameTest(unittest.TestCase):

    def testCorrectNaming(self):
        class Test(XMLObject):
            _name = 'altTest'
        self.assertEqual(Test().toXml(headers=0), "<altTest/>")

    def testSpaceNaming(self):
        class Test(XMLObject):
            _name = 'alt test'
        self.assertEqual(Test().toXml(headers=0), '<alt_test/>')

    def testTypeName(self):
        class IntTest(XMLObject):
            _name = 2

        class BoolTest(XMLObject):
            _name = True

        self.assertRaises(ValueError, IntTest().toXml)
        self.assertRaises(ValueError, IntTest().toXml)

class ImportExportTest(GetSetTest):

    def testImport(self):
        obj2 = Item.fromXml(self.xmlData)
        self.assertEqual(obj2, self.obj)
        
    def testExport(self):
        xml = self.obj.toXml(headers=0)
        self.assertEqual(xml, self.xmlData)
        
    def testExportImport(self):
        obj2 = Item.fromXml(self.obj.toXml())
        self.assertEqual(obj2, self.obj)

class InheritanceTest(unittest.TestCase):

    def testSimpleInherit(self):

        class MyPlaylist(Playlist):
            description = TextNode()

        data = {'name':'test', 'type':'xml',
                'description': "This is the playlist description"}
        pl = MyPlaylist(**data)
        for k,v in data.iteritems():
            self.assertEqual(getattr(pl, k), v)
        self.assertEqual(pl.items, [])

    def testMultipleInherit(self):
        """ Damn this one is too crapy. Can't use assertRaises though
            because the exception is raised during class creatioin
        """
        class Dummy:
            _name = 'foo bar'

        try:
            class MyOtherPlaylist(Playlist, Dummy):
                pass
        except Exception,ex:
            self.assertEqual(str(ex), "Multiple inheritance is EVIL. Don't use it")


class Item2(XMLObject):
    pass

class DummyItem(XMLObject):
    bl = ItemNode('Blah')
    
class Blah(XMLObject):
    mix = ChoiceNode(['Item2','Blah'])

class TestChoice(unittest.TestCase):

    def testLeftRecurse(self):
        class Item2(XMLObject):
            bl = ItemNode('Blah')
            txt = TextNode()
        b = Blah()
        it = Item2(txt="glop")
        it2 = DummyItem()
        self.assertRaises(ValueError, setattr, b, 'mix', it)
        self.assertRaises(ValueError, setattr, b, 'mix', it2)
        
    def testLeftRecurse2(self):
        b = Blah()
        b.mix = Item2(txt='glap')
        self.assertEqual(b.mix, Item2(txt='glap'))
                         
        class Blah2(XMLObject):
            mix = ChoiceNode(['Blah2', 'Item2'])

        self.assertRaises(ValueError, Blah2)

    def testMultiples(self):
        class Item(XMLObject):
            content = RawNode()

        class Blah(XMLObject):
            mix = ChoiceNode(['Item', 'Blah'], noLimit=True)

        it1 = Item(content='item1')
        b1 = Blah()
        b1.mix.append(it1)
        b1.mix.append(b1)
        b1.mix.append(Item(content='item2'))
        b1.mix.append(it1)
        b1.mix.append(b1)


    def testOnce(self):
        class Item(XMLObject):
            content = RawNode()

        class Blah(XMLObject):
            mix = ChoiceNode(['Item', 'Blah'], noLimit=False)

        it1 = Item(content='item1')
        b1 = Blah()
        b1.mix = it1

        b2 = Blah.fromXml(b1.toXml())
        self.assertEqual(b1,b2)

class XXX(XMLObject):
    """
    <!ELEMENT XXX (AAA+ , BBB+)>
    <!ELEMENT AAA (BBB | CCC )>
    <!ELEMENT BBB (#PCDATA | CCC )*>
    <!ELEMENT CCC (#PCDATA)>
    """
    aaa = ListNode('AAA',optional=False)
    bbb = ListNode('BBB',optional=False)
    
class AAA(XMLObject):
    mainNode = ChoiceNode(['BBB','CCC'])

class BBB(XMLObject):
    mainNode = ChoiceNode(['#PCDATA', 'CCC'],optional=True, noLimit=True)

class CCC(XMLObject):
    mainNode = TextNode()


class TestChoiceListRaw(unittest.TestCase):

    def setUp(self):
        self.xxx = XXX()
        self.xxx.aaa.append(CCC('Precisely one element.'))

        bb = BBB()
        bb.append(CCC())
        bb.append(CCC())
        bb.append(CCC())

        self.xxx.aaa.append(bb)
        self.xxx.bbb.append(BBB())

        bb2 = BBB()
        bb2.append('This is ')
        bb2.append(CCC())
        bb2.append(' a combination ')
        bb2.append(CCC())
        bb2.append(' of ')
        bb2.append(CCC('CCC elements'))
        bb2.append('and text')
        bb2.append(CCC())

        self.xxx.bbb.append(bb2)
        self.xxx.bbb.append(BBB('Text only.'))

    def testExportImport(self):
        xxxStr = self.xxx.toXml()
        xxx2 = XXX.fromXml(xxxStr)
        self.assertEqual(xxx2, self.xxx)

    def testXXX(self):
        xx = XXX()
        self.assertRaises(TypeError, apply, 'append', xx.aaa, 'Some Text')
        self.assertRaises(TypeError, apply, 'toXml',  xx)

    def testAAA(self):
        aa = AAA()
        self.assertRaises(TypeError, apply, 'toXml',  aa)
        self.assertRaises(TypeError, AAA, 'Blah')
        aa = AAA(BBB('Test'))
        self.assertEqual(aa.mainNode, BBB('Test'))
        aa = AAA(CCC('Test2'))
        self.assertEqual(aa.mainNode, CCC('Test2'))

    def testBBB(self):
        bb = BBB()
        self.assertEqual(bb.toXml(headers=0), '<bBB/>')
        bb = BBB()
        bb.append('Toto')
        bb.append(CCC('Foo'))
        self.assertEqual(len(bb.mainNode), 2)

class EntitiesTest(unittest.TestCase):

    def testQuote(self):
        class QuoteTest(XMLObject):
            _entities = [ ('&foo;', 'Bar')]
            title = StringAttribute()

        q = QuoteTest(title='some " quote &foo;')
        expected = '<quoteTest title="some &quot; quote Bar"/>'
        self.assertEqual(q.toXml(headers=0), expected)
        self.assertEqual(QuoteTest.fromXml(expected).title, 'some " quote Bar')

    def testText(self):
        class Text(XMLObject):
            mainNode = RawNode()

        t = Text("""Some ' and may be "some" quotes and &""")
        expected = """\
<text>
    <![CDATA[Some ' and may be "some" quotes and &]]>
</text>"""
        self.assertEqual(t.toXml(headers=0), expected)


class A(XMLObject):
    name = TextNode()
    opt_param = TextNode(optional = True)

class B(XMLObject):
    name = TextNode()
    a_list = ListNode("A")

class TobiasTest(unittest.TestCase):
    def testNameClash(self):               
        b = B(name = "BLAH")
        b.a_list.append(A(name="1"))
        b.a_list.append(A(name="2"))

        xml = b.toXml()
        b1 = B.fromXml(xml)

        self.assertEqual(b1.name , "BLAH")
        self.assertEqual(len(b1.a_list) ,2)

    def testNoneAttr(self):               
        a1 = A(name = "foo")
        self.assertEqual(a1.opt_param, None)
        a2 = A().fromXml(a1.toXml())
        self.assertEqual(a2.opt_param, None)


class TwoDeep(XMLObject):
    pass

class OneDeep(XMLObject):
    children = ListNode('TwoDeep')
    
class Root(XMLObject):
    children = ListNode('OneDeep')

class DeepTest(unittest.TestCase):
    expected = """\
<root>
  <oneDeep>
    <twoDeep/>
  </oneDeep>
</root>"""

    def testOneAfterTwo(self):
        root = Root()
        one = OneDeep()
        two = TwoDeep()
        one.children.append(two)
        root.children.append(one)
        self.assertEqual(str(root), self.expected)
        self.assertEqual(str(Root.fromXml(str(root))),
                         self.expected)
            
    def testOneBeforeTwo(self):
        root = Root()
        one = OneDeep()
        root.children.append(one)
        two = TwoDeep()
        one.children.append(two)
        self.assertEqual(str(root),self.expected)
        self.assertEqual(str(Root.fromXml(str(root))),
                         self.expected)


    def testFirstChild(self):
        root = Root()
        one = OneDeep()
        root.children.append(one)
        two = TwoDeep()
        one.children.append(two)
        self.assertEqual(root.children[0],one)

        
if __name__ == '__main__':
    unittest.main()
